use std::collections::HashMap;

use crate::{
    error,
    request::Request,
    util::{self, ErrorKind},
};
use colored::*;
use reqwest::StatusCode;
use serde_derive::{Deserialize, Serialize};

pub struct User;

#[derive(Deserialize)]
pub struct MultipleUsers {
    pub data: Vec<UserResponse>,
    pub ok: bool,
}

#[derive(Serialize, Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct UserResponse {
    #[serde(rename = "avatar_url")]
    pub avatar_url: String,
    pub created: String,
    pub email: String,
    #[serde(rename = "full_name")]
    pub full_name: String,
    pub id: i64,
    #[serde(rename = "is_admin")]
    pub is_admin: bool,
    pub language: String,
    #[serde(rename = "last_login")]
    pub last_login: String,
    pub login: String,
    pub restricted: bool,
}

#[derive(Deserialize)]
#[serde(rename_all = "camelCase")]
pub struct UserErrorResponse {
    pub message: String,
    pub url: String,
}

impl User {
    pub fn new() -> User {
        User {}
    }

    pub fn create_user(&self, request: &Request) -> Result<(), error::ErrorKind> {
        let client = &request.client;
        let mut user_create_input: HashMap<String, String> = HashMap::new();
        let url = format!(
            "{request}/admin/users?token={api_token}",
            request = request.url.as_ref().unwrap(),
            api_token = request.authentication.credentials.1.as_ref().unwrap()
        );

        let username = util::get_input(String::from("Please enter a username:"));
        let email = util::get_input(String::from("Please enter a email address:"));
        let password = util::get_input(String::from("Please enter a password:"));

        user_create_input.insert("login".to_string(), username);
        user_create_input.insert("email".to_string(), email);
        user_create_input.insert("password".to_string(), password);

        let response = client.post(url.as_str()).json(&user_create_input).send();
        // TODO: match return type
        match response {
            Ok(repo) => match repo.status() {
                StatusCode::CREATED => {
                    let deserialized: UserResponse = repo.json().unwrap();
                    println!("{}", "User successfully created!".green());
                    println!(
                        "\tFull Name: {:0}\n\tUsername: {:1}\n\tEmail: {:2}\n",
                        deserialized.full_name, deserialized.login, deserialized.email
                    );
                    Ok(())
                }
                StatusCode::BAD_REQUEST => {
                    let deserialized: UserErrorResponse = repo.json().unwrap();
                    Err(error::ErrorKind::BadRequest(deserialized.message))
                }
                StatusCode::FORBIDDEN => {
                    let deserialized: UserErrorResponse = repo.json().unwrap();
                    Err(error::ErrorKind::ForbiddenRequest(deserialized.message))
                }
                _ => Err(error::ErrorKind::Other),
            },
            Err(_e) => Err(error::ErrorKind::Other),
        }
    }

    pub fn list_user(&self, request: &Request) -> Result<(), error::ErrorKind> {
        let client = &request.client;

        let url = format!(
            "{request}/users?token={api_token}",
            request = request.url.as_ref().unwrap(),
            api_token = request.authentication.credentials.1.as_ref().unwrap(),
        );
        let response = client.get(url).send();
        match response {
            Ok(repo) => {
                match repo.status() {
                    StatusCode::OK => {
                        let deserialized: MultipleUsers = repo.json().unwrap();

                        match deserialized.data.len() != 0 {
                            true => {
                                println!("{}", "List of users found:");

                                for (i, data) in deserialized.data.iter().enumerate() {
                                    println!("{}.\nUsername: {:1}\n\tEmail address: {:2}\n\tAdmin?: {:3}", i + 1, data.login, data.email, data.is_admin);
                                }

                                println!(
                                    "Total number of users indexed: {}",
                                    deserialized.data.len()
                                );
                                Ok(())
                            }
                            false => {
                                println!("{}", "No users found :(");
                                Ok(())
                            }
                        }
                    }
                    StatusCode::BAD_REQUEST => {
                        let deserialized: UserErrorResponse = repo.json().unwrap(); // TODO: handle
                        Err(error::ErrorKind::BadRequest(deserialized.message))
                    }
                    StatusCode::FORBIDDEN => {
                        let deserialzed: UserErrorResponse = repo.json().unwrap(); // TODO: handle errs
                        Err(error::ErrorKind::ForbiddenRequest(deserialzed.message))
                    }
                    _ => Err(error::ErrorKind::Other),
                }
            }
            Err(_e) => Err(error::ErrorKind::Other),
        }
    }

    pub fn search_user(&self, request: &Request) -> Result<(), error::ErrorKind> {
        let client = &request.client;
        let arg_value = request
            .arg_value
            .subcommand()
            .1
            .unwrap()
            .value_of("search")
            .unwrap();

        let url = format!(
            "{request}/users?q={query}&token={api_token}",
            request = request.url.as_ref().unwrap(),
            query = arg_value,
            api_token = request.authentication.credentials.1.as_ref().unwrap(),
        );

        let response = client.get(url.as_str()).send();

        match response {
            Ok(repo) => match repo.status() {
                StatusCode::OK => Ok(()),
                _ => Err(error::ErrorKind::Other),
            },
            Err(_e) => Err(error::ErrorKind::Other),
        }
    }

    pub fn delete_user(&self, request: &Request) -> Result<(), error::ErrorKind> {
        let client = &request.client;
        let arg_value = request
            .arg_value
            .subcommand()
            .1
            .unwrap()
            .value_of("search")
            .unwrap();

        let url = format!(
            "{request}/admin/users/{username}",
            request = request.url.as_ref().unwrap(),
            username = arg_value,
        );
        let response = client.delete(url.as_str()).send();

        match response {
            Ok(repo) => match repo.status() {
                StatusCode::NO_CONTENT => {
                    println!("User successfully deleted!");
                    Ok(())
                }
                StatusCode::FORBIDDEN => {
                    let deserialized: UserErrorResponse = repo.json().unwrap();
                    Err(error::ErrorKind::ForbiddenRequest(deserialized.message))
                }
                _ => Err(error::ErrorKind::Other),
            },
            Err(_e) => Err(error::ErrorKind::Other),
        }
    }

    pub fn authed_user(&self, request: &Request) {
        let client = &request.client;
    }
}

impl Default for User {
    fn default() -> Self {
        User::new()
    }
}
